/**

Copyright (c) Scott Gasch

Module Name:

    util.c

Abstract:

    General purpose utility code to support the engine.

Author:

    Scott Gasch (scott.gasch@gmail.com) 8 Apr 2004

Revision History:

    $Id: util.c 355 2008-07-01 15:46:43Z scott $

**/

#include "chess.h"
#include <errno.h>

extern uint32_t g_uIterateDepth;
extern uint32_t g_uFullWidthDepth;

#define STATE_UNKNOWN  (0)
#define STATE_HEADER   (1)
#define STATE_MOVELIST (2)

int8_t *
ReadNextGameFromPgnFile(FILE *pf)
/**

Routine description:

    Reads the next game in a PGN file and returns a pointer to it.

Parameters:

    FILE *pf : open file pointer

Return value:

    int8_t * : NULL on error, else caller must free via SystemFreeMemory.

**/
{
    int8_t buf[MEDIUM_STRING_LEN_CHAR];
    int8_t *p;
    int32_t lStartOfGame = -1;
    int32_t lEndOfGame = -1;
    int32_t lBefore;
    uint32_t uBytes;
    uint32_t uState = STATE_UNKNOWN;

    while(1)
    {
        lBefore = ftell(pf);
        if (fgets(buf, ARRAY_LENGTH(buf), pf) <= 0)
        {
            //Trace("fgets failed, errno=%u.\n", errno);
            break;
        }

        //
        // Skip leading whitespace on this line...
        //
        p = buf;
        while(*p && isspace(*p)) p++;

        if (*p == '\0')
        {
            //
            // We just read a blank line
            //
            switch(uState)
            {
                case STATE_UNKNOWN:
                    break;

                case STATE_HEADER:
                    uState = STATE_MOVELIST;
                    break;

                case STATE_MOVELIST:
                    lEndOfGame = lBefore;
                    uBytes = lEndOfGame - lStartOfGame;
                    if (uBytes <= 0)
                    {
                        return(NULL);
                    }

                    p = SystemAllocateMemory(uBytes + 1);
                    ASSERT(p);
                    fseek(pf, lStartOfGame, SEEK_SET);
                    if (1 != fread(p, uBytes, 1, pf))
                    {
                        SystemFreeMemory(p);
                        return(NULL);
                    }
                    p[uBytes] = '\0';
                    return(p);
            }
        }
        else
        {
            //
            // We saw a non-blank line
            //
            switch(uState)
            {
                case STATE_HEADER:
                case STATE_MOVELIST:
                    break;

                case STATE_UNKNOWN:
                    if (*p == '[')
                    {
                        uState = STATE_HEADER;
                        lStartOfGame = lBefore;
                    }
                    break;
            }
        }
    }
    return(NULL);
}


COMMAND(LearnPsqtFromPgn)
/**

Routine description:

Parameters:

    int8_t *szFilename,

Return value:

    FLAG

**/
{
    int8_t *szFilename;
    uint32_t uPSQT[12][128];
    int8_t *pPGN;
    FILE *pf = NULL;
    DLIST_ENTRY *p;
    GAME_MOVE *q;
    uint32_t x, y, z;
    double d;
    struct stat s;
#ifdef DEBUG
    uint32_t uHeap;
#endif

    if (argc < 2)
    {
        Trace("Error (missing filename)\n");
        return;
    }
    szFilename = argv[1];

    if (FALSE == SystemDoesFileExist(szFilename))
    {
        Trace("LearnPsqtFromPgn: file \"%s\" doesn't exist.\n",
              szFilename);
        return;
    }

    if (0 != stat(szFilename, &s))
    {
        Trace("Stat failed.\n");
        return;
    }

    pf = fopen(szFilename, "rb");
    if (NULL == pf)
    {
        Trace("LearnPsqtFromPgn: error reading file \"%s\".\n",
              szFilename);
        return;
    }

    memset(uPSQT, 0, sizeof(uPSQT));
#ifdef DEBUG
    ResetGameList();
    uHeap = GetHeapMemoryUsage();
#endif
    while((pPGN = ReadNextGameFromPgnFile(pf)) != NULL)
    {
        d = (double)ftell(pf);
        d /= (double)s.st_size;
        d *= 100.0;
        printf("(%5.2f%% done)\r", d);

        if (TRUE == LoadPgn(pPGN))
        {
            p = g_GameData.sMoveList.pFlink;
            while(p != &(g_GameData.sMoveList))
            {
                q = CONTAINING_STRUCT(p, GAME_MOVE, links);
                if ((q->uNumber > 10) &&
                    (q->uNumber < 50) &&
                    (!q->mv.pCaptured))
                {
                    uPSQT[q->mv.pMoved - 2][q->mv.cTo] += 1;
                }
                p = p->pFlink;
            }
        }
        SystemFreeMemory(pPGN);
#ifdef DEBUG
        ResetGameList();
        ASSERT(uHeap == GetHeapMemoryUsage());
#endif
    }
    fclose(pf);

    z = 0;
    for (x = 0; x < 12; x++)
    {
        for (y = 0; y < 128; y++)
        {
            if (uPSQT[x][y] > z)
            {
                z = uPSQT[x][y];
            }
        }
    }

    Trace("uint32_t g_PSQT[12][128] = {\n");
    for (x = 0; x < 12; x++)
    {
        Trace("// Piece 0x%x (%s)\n", x + 2, PieceAbbrev((PIECE)(x + 2)));
        Trace("{\n");

        for (y = 0; y < 128; y++)
        {
            if (y % 16 == 0) Trace("\n");
            d = (double)uPSQT[x][y];
            d /= (double)z;
            d *= (double)1000.0;
            Trace("%4u,", (uint32_t)d);
        }
        Trace("},\n");
    }
    Trace("};\n");
}


COMMAND(GeneratePositionAndBestMoveSuite)
/**

Routine description:

Parameters:

    int8_t *szFilename,

Return value:

    FLAG

**/
{
    int8_t *szFilename;
    int8_t *pPGN;
    FILE *pf = NULL;
    DLIST_ENTRY *p;
    GAME_MOVE *q;
    SEARCHER_THREAD_CONTEXT *ctx = NULL;
    FLAG fPost = g_Options.fShouldPost;
    struct stat s;
    double d;
    POSITION board;

    if (argc < 2)
    {
        Trace("Error (missing filename)\n");
        goto end;
    }
    szFilename = argv[1];

    if (FALSE == SystemDoesFileExist(szFilename))
    {
        Trace("LearnPsqtFromPgn: file \"%s\" doesn't exist.\n",
              szFilename);
        goto end;
    }

    if (0 != stat(szFilename, &s))
    {
        Trace("Stat failed.\n");
        goto end;
    }

    pf = fopen(szFilename, "rb");
    if (NULL == pf)
    {
        Trace("LearnPsqtFromPgn: error reading file \"%s\".\n",
              szFilename);
        goto end;
    }

    ctx = SystemAllocateMemory(sizeof(SEARCHER_THREAD_CONTEXT));
    if (NULL == ctx)
    {
        Trace("Out of memory.\n");
        goto end;
    }

    g_Options.fShouldPost = FALSE;
    while((pPGN = ReadNextGameFromPgnFile(pf)) != NULL)
    {
        d = (double)ftell(pf);
        d /= (double)s.st_size;
        d *= 100.0;
        printf("(%5.2f%% done)\r", d);

        if (TRUE == LoadPgn(pPGN))
        {
            p = g_GameData.sMoveList.pFlink;
            while(p != &(g_GameData.sMoveList))
            {
                q = CONTAINING_STRUCT(p, GAME_MOVE, links);
                if ((q->uNumber > 20) && (q->uNumber < 60))
                {
                    if (FenToPosition(&board, q->szUndoPositionFen))
                    {
                        InitializeSearcherContext(&board, ctx);
                        g_MoveTimer.bvFlags = 0;
                        g_Options.fPondering = FALSE;
                        g_Options.fThinking = TRUE;
                        g_Options.fSuccessfulPonder = FALSE;

                        MaintainDynamicMoveOrdering();
                        DirtyHashTable();

                        g_MoveTimer.uNodeCheckMask = 0x1000 - 1;
                        g_MoveTimer.dStartTime = SystemTimeStamp();
                        g_MoveTimer.bvFlags = 0;
                        g_MoveTimer.dSoftTimeLimit =
                            g_MoveTimer.dStartTime + 3.0;
                        g_MoveTimer.dHardTimeLimit =
                            g_MoveTimer.dStartTime + 3.0;
                        g_Options.uMaxDepth = MAX_DEPTH_PER_SEARCH - 1;

                        //
                        // TODO: Set draw value
                        //
#if (PERF_COUNTERS && MP)
                        ClearHelperThreadIdleness();
#endif
                        GAME_RESULT result = Iterate(ctx);
                        if (result.eResult == RESULT_IN_PROGRESS) {
                            ASSERT(SanityCheckMove(&board,
                                                   ctx->sPlyInfo[0].PV[0]));
                            Trace("setboard %s\n", q->szUndoPositionFen);
                            Trace("gmmove %s\n", q->szMoveInSan);
                            Trace("solution %s\n",
                                  MoveToSan(ctx->sPlyInfo[0].PV[0], &board));
                        }
                    }
                }
                p = p->pFlink;
            }
        }
        SystemFreeMemory(pPGN);
    }

 end:
    if (NULL != pf)
    {
        fclose(pf);
    }

    if (NULL != ctx)
    {
        SystemFreeMemory(ctx);
    }

    g_Options.fShouldPost = fPost;
}


int8_t *
FindChunk(int8_t *sz, uint32_t uTargetChunk)
/**

Routine description:

    A support routine used by FenToPosition.  Used to parse apart a
    string and return the Nth chunk of text (delimited by whitespace)

Parameters:

    int8_t *sz,
    uint32_t uTargetChunk

Return value:

    int8_t *

**/
{
    int8_t *p = sz;
    uint32_t u = 0;

    if (0 == uTargetChunk) return(p);

    do
    {
        //
        // Skip whitespace between chunks
        //
        while(isspace(*p) && *p) p++;
        if (!*p) return(NULL);

        u++;
        if (u == uTargetChunk) return(p);

        //
        // Skip chunks
        //
        while(!isspace(*p) && &p) p++;
        if (!*p) return(NULL);
    }
    while(1);

#ifdef DEBUG
    UtilPanic(SHOULD_NOT_GET_HERE,
              NULL, NULL, NULL, NULL,
              __FILE__, __LINE__);
#endif
    return(NULL);
}



int8_t *
ScoreToString(SCORE iScore)
/**

Routine description:

    Format a chess score as a string that looks like:

        +0.100
        -4.951
        +MATE20
        -MATE4
        +0.000

    etc...

    Note: NOT thread safe.

Parameters:

    SCORE iScore

Return value:

    int8_t *

**/
{
    static int8_t szBuf[10];
    uint32_t uDist;

    if (abs(iScore) < +NMATE)
    {
        snprintf(szBuf, ARRAY_LENGTH(szBuf),
                 "%c%u.%02u",
                 (iScore < 0) ? '-' : '+',
                 abs(iScore / 100),
                 abs(iScore % 100));
        return(szBuf);
    }

    if (iScore > 0)
    {
        uDist = +INFINITY - iScore;
        uDist /= 2;
        if (uDist)
        {
            snprintf(szBuf, ARRAY_LENGTH(szBuf), "+MATE%u", uDist);
        }
        else
        {
            snprintf(szBuf, ARRAY_LENGTH(szBuf), "+MATE");
            Trace("computer mates\n");


        }
    }
    else
    {
        uDist = +INFINITY + iScore;
        uDist /= 2;
        if (uDist)
        {
            snprintf(szBuf, ARRAY_LENGTH(szBuf), "-MATE%u", uDist);
        }
        else
        {
            snprintf(szBuf, ARRAY_LENGTH(szBuf), "-MATE");
             Trace("computer mates\n");


        }
    }
    return(szBuf);
}



int8_t *
CoorToString(COOR c)
/**

Routine description:

    Take a chessboard coordinate and convert it into a string.  Note:
    NOT thread safe.

Parameters:

    COOR c

Return value:

    int8_t *

**/
{
    static int8_t buf[3];

    if (IS_ON_BOARD(c))
    {
        buf[0] = 'a' + MYFILE(c);
        buf[1] = '0' + RANK(c);
        buf[2] = '\0';
    }
    else
    {
        buf[0] = '-';
        buf[1] = '\0';
    }
    return(buf);
}


int8_t *
ColorToString(uint32_t u)
/**

Routine description:

    Convert a color into a string ("WHITE" or "BLACK").  Callers
    should not mess with the contents of the buffer returned!

Parameters:

    uint32_t u

Return value:

    int8_t

**/
{
    static int8_t *szColor[2] =
    {
        "BLACK",
        "WHITE"
    };
    ASSERT(IS_VALID_COLOR(u));
    return(szColor[u]);
}


int8_t *
TimeToString(double d)
/**

Routine description:

    Convert a time value into a string.  e.g.

        00:00:01.021                          // seconds with ms
        00:15:21.998                          // minutes+seconds
        01:59:23.232                          // hours+minutes+sec
        27:08:11:103                          // we don't do days

    Note: NOT thread safe.  The caller should not mess with the
    buffer returned.

Parameters:

    double d

Return value:

    int8_t *

**/
{
    uint32_t uTotalSecs = (uint32_t)d;
    double dFract = d - (double)uTotalSecs;
    uint32_t uHours;
    uint32_t uMinutes;
    uint32_t uFract;
    static int8_t buf[13];

    dFract *= 1000;
    uFract = (int32_t)dFract;
    uHours = uTotalSecs / 60 / 60;
    uTotalSecs -= uHours * 60 * 60;
    uMinutes = uTotalSecs / 60;
    uTotalSecs -= uMinutes * 60;
    ASSERT(uTotalSecs < 60);

    if (uHours != 0)
    {
        snprintf(buf, ARRAY_LENGTH(buf),
                 "%2u:%02u:%02u.%03u",
                 uHours, uMinutes, uTotalSecs, uFract);
    }
    else
    {
        snprintf(buf, ARRAY_LENGTH(buf),
                 "%02u:%02u.%03u",
                 uMinutes, uTotalSecs, uFract);
    }
    return(buf);
}


FLAG
BufferIsZeroed(uint8_t *p, uint32_t u)
/**

Routine description:

    Verify that a buffer is zeroed out.

Parameters:

    uint8_t *p : buffer start
    uint32_t u : buffer length

Return value:

    FLAG

**/
{
    while(u)
    {
        if (*p != (uint8_t)0)
        {
            return(FALSE);
        }
        p++;
        u--;
    }
    return(TRUE);
}


void CDECL
Trace(int8_t *szMessage, ...)
/**

Routine description:

    Trace a logging message on stdout and to the logfile

Parameters:

    int8_t *szMessage,
    ...

Return value:

    void

**/
{
    va_list ap;                            // Variable argument list
    int8_t buf[MEDIUM_STRING_LEN_CHAR];

    memset(buf, 0, sizeof(buf));

    //
    // Populate buffer
    //
    va_start(ap, szMessage);
    COMPILER_VSNPRINTF(buf, MEDIUM_STRING_LEN_CHAR - 1, szMessage, ap);
    va_end(ap);

    //
    // Send to logfile
    //
    if (NULL != g_pfLogfile) {
        fprintf(g_pfLogfile, buf);
    }

    //
    // Send to stdout
    //
    fprintf(stdout, buf);
    fflush(stdout);
}


void CDECL
Log(int8_t *szMessage, ...)
/**

Routine description:

    Trace a logging message to the logfile only

Parameters:

    int8_t *szMessage,
    ...

Return value:

    void

**/
{
    va_list ap;                            // Variable argument list
    int8_t buf[MEDIUM_STRING_LEN_CHAR];
    memset(buf, 0, sizeof(buf));

    //
    // Populate buffer
    //
    va_start(ap, szMessage);
    COMPILER_VSNPRINTF(buf, MEDIUM_STRING_LEN_CHAR - 1, szMessage, ap);
    va_end(ap);

    //
    // Send to logfile
    //
    if (NULL != g_pfLogfile) {
        fprintf(g_pfLogfile, buf);
        fflush(g_pfLogfile);
    }
}


void CDECL
Bug(int8_t *szMessage, ...)
/**

Routine description:

    Trace an error message to stderr and the logfile

Parameters:

    int8_t *szMessage,
    ...

Return value:

    void

**/
{
    va_list ap;                            // Variable argument list
    int8_t buf[MEDIUM_STRING_LEN_CHAR];

    memset(buf, 0, sizeof(buf));

    //
    // Populate buffer
    //
    va_start(ap, szMessage);
    COMPILER_VSNPRINTF(buf, MEDIUM_STRING_LEN_CHAR - 1, szMessage, ap);
    va_end(ap);

    //
    // Send it to logfile
    //
    if (NULL != g_pfLogfile) {
        fprintf(g_pfLogfile, buf);
        fflush(g_pfLogfile);
    }

    //
    // Send it to stderr
    //
    fprintf(stderr, buf);
    fflush(stderr);
}


#ifdef DEBUG
void
_assert(int8_t *szFile, uint32_t uLine)
/**

Routine description:

    An assertion handler

Parameters:

    int8_t *szFile,
    uint32_t uLine

Return value:

    void

**/
{
    Bug("Assertion failure in %s, line %u.\n", szFile, uLine);
    BREAKPOINT;
}
#endif


int32_t
AcquireSpinLock(volatile int32_t *pSpinlock)
/**

Routine description:

    Atomically acquire a lock or spin trying.

Parameters:

    uint32_t *pSpinlock : a pointer to the lock

Return value:

    uint32_t : the number of times we had to spin

**/
{
    uint32_t uLoops = 0;
    while(0 != InterlockedCompareExchange(pSpinlock, 1, 0))
    {
        uLoops++;
    }
    ASSERT(*pSpinlock == 1);
    return(uLoops);
}

FLAG
TryAcquireSpinLock(volatile int32_t *pSpinlock)
{
    return(0 != InterlockedCompareExchange(pSpinlock, 1, 0));
}

void
ReleaseSpinLock(volatile int32_t *pSpinlock)
/**

Routine description:

    Release a lock so someone else can acquire it.

Parameters:

    uint32_t *pSpinlock : a pointer to the lock

Return value:

    void

**/
{
    uint32_t u;
    ASSERT(*pSpinlock == 1);
    u = InterlockedCompareExchange(pSpinlock, 0, 1);
    ASSERT(u == 1);
}


FLAG
BackupFile(int8_t *szFile)
/**

Routine description:

    Backup a file to file.000.  If file.000 already exists, back it up
    to file.001 etc...

    Note: this function is recursive and can require quite a lot of
    stack space.  Also it is full of race conditions and should not be
    used when some other code may be accessing the filesystem at the
    same time.

Parameters:

    int8_t *szFile : the file to backup

Return value:

    FLAG

**/
{
    uint32_t u;
    int8_t buf[SMALL_STRING_LEN_CHAR];
    int8_t *p;

    if (TRUE == SystemDoesFileExist(szFile))
    {
        if ((strlen(szFile) > 3) &&
            (isdigit(szFile[0])) &&
            (isdigit(szFile[1])) &&
            (isdigit(szFile[2])))
        {
            u = (szFile[0] - '0') * 100;
            u += (szFile[1] - '0') * 10;
            u += szFile[2] - '0' + 1;
            p = &(szFile[3]);
        }
        else
        {
            u = 0;
            p = szFile;
        }
        snprintf(buf, ARRAY_LENGTH(buf), "%03u%s", u, p);
        if (TRUE == SystemDoesFileExist(buf))
        {
            BackupFile(buf);
        }
        return(SystemCopyFile(szFile, buf) && SystemDeleteFile(szFile));
    }
    return(FALSE);
}


uint8_t
Checksum(uint8_t *p, uint32_t uSize)
/**

Routine description:

    Compute the checksum of a buffer.

Parameters:

    uint8_t *p,
    uint32_t uSize

Return value:

    uint8_t

**/
{
    uint8_t b = 0;
    while(uSize)
    {
        b ^= *p;
        p++;
        uSize--;
    }
    return(b);
}

void
FinishPVTailFromHash(SEARCHER_THREAD_CONTEXT *ctx,
                     int8_t *buf,
                     uint32_t uLenRemain)
{
    MOVE PV[MAX_PLY_PER_SEARCH];
    MOVE mv;
    uint32_t uPly = ctx->uPly;
    uint32_t uLen;
#ifdef DEBUG
    POSITION board;
    memcpy(&board, &(ctx->sPosition), sizeof(POSITION));
#endif

    if (NULL == g_pHashTable) return;
    do
    {
        mv = GetPonderMove(&ctx->sPosition);
        if (mv.uMove == 0) break;
        PV[ctx->uPly] = mv;

        uLen = (uint32_t)strlen(MoveToSan(mv, &ctx->sPosition)) + 1;
        ASSERT(uLen <= 11);
        if (uLenRemain <= uLen) break;
        strcat(buf, MoveToSan(PV[ctx->uPly], &ctx->sPosition));
        strcat(buf, " ");
        uLenRemain -= uLen;

        if (FALSE == MakeMove(ctx, mv)) break;
    }
    while(1);

    //
    // Unmake the moves
    //
    while(ctx->uPly > uPly)
    {
        UnmakeMove(ctx, PV[ctx->uPly - 1]);
    }
    ASSERT(ctx->uPly == uPly);
    ASSERT(PositionsAreEquivalent(&board, &ctx->sPosition));
}



int8_t *
WalkPV(SEARCHER_THREAD_CONTEXT *ctx)
/**

Routine description:

    Return a string buffer representing the principle variation.

Parameters:

    SEARCHER_THREAD_CONTEXT *ctx

Return value:

    int8_t *

**/
{
    static int8_t buf[SMALL_STRING_LEN_CHAR];
    uint32_t u = ctx->uPly;
    uint32_t v = u;
    MOVE mv;
#ifdef DEBUG
    POSITION board;
    memcpy(&board, &(ctx->sPosition), sizeof(POSITION));
#endif

    buf[0] = '\0';
    while((mv.uMove = ctx->sPlyInfo[v].PV[u].uMove) != 0)
    {
        if (mv.uMove == HASHMOVE.uMove)
        {
            strncat(buf, "<TT> ", ARRAY_LENGTH(buf) - strlen(buf) - 1);
            FinishPVTailFromHash(ctx, buf, ARRAY_LENGTH(buf) - strlen(buf));
            ASSERT(ctx->sPlyInfo[v].PV[u+1].uMove == 0);
            break;
        }
        else if (mv.uMove == RECOGNMOVE.uMove)
        {
            strncat(buf, "<REC>", ARRAY_LENGTH(buf) - strlen(buf) - 1);
            ASSERT(ctx->sPlyInfo[v].PV[u+1].uMove == 0);
            break;
        }
        else if (mv.uMove == DRAWMOVE.uMove)
        {
            strncat(buf, "<=>", ARRAY_LENGTH(buf) - strlen(buf) - 1);
            ASSERT(ctx->sPlyInfo[v].PV[u+1].uMove == 0);
            break;
        }
        else
        {
            strncat(buf, MoveToSan(mv, &ctx->sPosition),
                    ARRAY_LENGTH(buf) - strlen(buf));
            strncat(buf, " ", ARRAY_LENGTH(buf) - strlen(buf) - 1);
        }

        if (FALSE == MakeMove(ctx, mv)) break;
        u++;

        if (strlen(buf) > (SMALL_STRING_LEN_CHAR - 10)) break;
    }

    //
    // Unmake the moves
    //
    u--;
    while(ctx->uPly > v)
    {
        UnmakeMove(ctx, ctx->sPlyInfo[v].PV[u]);
        u--;
    }
    ASSERT(ctx->uPly == v);
    ASSERT(PositionsAreEquivalent(&board, &ctx->sPosition));
    return(buf);
}


void
UtilPanic(uint32_t uPanicCode,
          POSITION *pos,
          void *arg1, void *arg2, void *arg3,
          int8_t *szFile, uint32_t uLine)
{
    Bug("Typhoon PANIC 0x%x (%p, %p, %p, %p)\n"
        "Location: %s at line %u\n\n",
        uPanicCode, pos, arg1, arg2, arg3, szFile, uLine);
    Banner();
    Bug("----------------------------------------------------------------\n");
    switch(uPanicCode) {
        case CANNOT_INITIALIZE_SPLIT:
        case GOT_ILLEGAL_MOVE_WHILE_PONDERING:
        case CANNOT_OFFICIALLY_MAKE_MOVE:
            DumpPosition(pos);
            DumpMove((uint32_t)arg1);
            break;
        case INITIALIZATION_FAILURE:
            Bug("%s\n", (int8_t *)arg1);
            break;
        default:
            break;
    }
    fflush(stdout);
    fflush(g_pfLogfile);
    BREAKPOINT;
    exit(-1);
}

void
UtilPrintPV(SEARCHER_THREAD_CONTEXT *ctx,
            SCORE iAlpha,
            SCORE iBeta,
            SCORE iScore,
            MOVE mv)
/**

Routine description:

    Print a PV out to stdout and the log.

Parameters:

    SEARCHER_THREAD_CONTEXT *ctx,
    SCORE iAlpha,
    SCORE iBeta,
    SCORE iScore,
    MOVE mv

Return value:

    void

**/
{
    double dNow = SystemTimeStamp();

    ASSERT(ctx->uThreadNumber == 0);
    ASSERT(IS_VALID_SCORE(iAlpha));
    ASSERT(IS_VALID_SCORE(iBeta));
    ASSERT(iAlpha < iBeta);
    ASSERT(IS_VALID_SCORE(iScore));

    dNow -= g_MoveTimer.dStartTime;

    //
    // Set the last PV in the context if we have a root PV or a root
    // fail high.
    //
    if ((iAlpha < iScore) && (iScore < iBeta))
    {
        strncpy(ctx->szLastPV, WalkPV(ctx), SMALL_STRING_LEN_CHAR);
    }
    else if (iScore > iBeta)
    {
        strncpy(ctx->szLastPV, MoveToSan(mv, &(ctx->sPosition)),
                SMALL_STRING_LEN_CHAR);
    }

    if ((TRUE == g_Options.fThinking) && (TRUE == g_Options.fShouldPost))
    {
        //
        // Maybe output the PV.  Note: xboard gets confused by PV
        // lines that don't match its requirements exactly; if we are
        // running under xboard, match its expected format.  Otherwise
        // be more helpful / verbose.
        //
        if ((dNow > 0.35) || ((dNow > 0.15) && (g_Options.fVerbosePosting)))
        {
            if (TRUE == g_Options.fRunningUnderXboard)
            {
                if ((iAlpha < iScore) && (iScore < iBeta))
                {
                    Trace("%2u %6d %5u %-12"
                          COMPILER_LONGLONG_UNSIGNED_FORMAT" %s\n",
                          g_uIterateDepth,
                          iScore,
                          (uint32_t)(dNow * 100),
                          ctx->sCounters.tree.u64TotalNodeCount,
                          ctx->szLastPV);
                }
                else if (iScore <= iAlpha)
                {
                    Trace("%2u %6d %5u %-12"
                          COMPILER_LONGLONG_UNSIGNED_FORMAT" FL --\n",
                          g_uIterateDepth,
                          iScore,
                          (uint32_t)(dNow * 100),
                          ctx->sCounters.tree.u64TotalNodeCount);
                }
                else
                {
                    ASSERT(iScore >= iBeta);
                    Trace("%2u %6d %5u %-12"
                          COMPILER_LONGLONG_UNSIGNED_FORMAT" %s ++\n",
                          g_uIterateDepth,
                          iScore,
                          (uint32_t)(dNow * 100),
                          ctx->sCounters.tree.u64TotalNodeCount,
                          ctx->szLastPV);
                }
            }
            else // !running under xboard
            {
                if ((iAlpha < iScore) && (iScore < iBeta))
                {
                    Trace("%2u  %5s  %-9s  %-11"
                          COMPILER_LONGLONG_UNSIGNED_FORMAT"  %s\n",
                          g_uIterateDepth,
                          ScoreToString(iScore),
                          TimeToString(dNow),
                          ctx->sCounters.tree.u64TotalNodeCount,
                          ctx->szLastPV);
                }
                else if (iScore <= iAlpha)
                {
                    Trace("%2u- %5s  %-9s  %-11"
                          COMPILER_LONGLONG_UNSIGNED_FORMAT"  FL --\n",
                          g_uIterateDepth,
                          ScoreToString(iScore),
                          TimeToString(dNow),
                          ctx->sCounters.tree.u64TotalNodeCount);
                }
                else
                {
                    ASSERT(iScore >= iBeta);
                    Trace("%2u+ %5s  %-9s  %-11"
                          COMPILER_LONGLONG_UNSIGNED_FORMAT"  %s ++\n",
                          g_uIterateDepth,
                          ScoreToString(iScore),
                          TimeToString(dNow),
                          ctx->sCounters.tree.u64TotalNodeCount,
                          ctx->szLastPV);
                }
            }
        }
    }
}
